import json
import subprocess
import tempfile
from pathlib import Path

from python_did_resolver.main import CardstackIdentifier

js_did_resolver_path = Path("../../packages/did-resolver").resolve()


def command(js_code):
    prefix = f"const o = require('{js_did_resolver_path}');"
    return [
        "node",
        "-r",
        "ts-node/register/transpile-only",
        "-e",
        prefix + js_code,
    ]


def run(js_code: str):
    try:
        x = subprocess.check_output(
            command(js_code), cwd=js_did_resolver_path, stderr=subprocess.STDOUT
        )
        o = json.loads(x)
        return o
    except subprocess.CalledProcessError as exc:
        print("Status : FAIL", exc.returncode, exc.output)


def js_resolve_did(did: str):
    res = run(
        f"""
        const {{ Resolver }} = require(\'did-resolver\');
        (async()=>{{
            const resolver = new Resolver({{
                ...o.getResolver(),
            }});
            did = o.encodeDID({{ type: \'PrepaidCardCustomization\'}})
            const a = await resolver.resolve(\"{did}\")
            console.log(JSON.stringify(a))
            }})();
        """,
    )
    return res


def js_encode_did(encode_options):
    tfile = tempfile.NamedTemporaryFile()
    with open(tfile.name, "w") as f:
        json.dump(encode_options, f)
    res = run(
        f"""
        const fs = require("fs")
        fs.readFile("{tfile.name}", "utf8",(err,data)=>{{
            let encodeOptions = JSON.parse(data)
            console.log(JSON.stringify(o.encodeDID(encodeOptions)))
        }}
        )
        """
    )
    return res


def js_parse_identifier(id: str) -> CardstackIdentifier:
    res = run(
        f"""
        console.log(JSON.stringify(o.parseIdentifier("{id}")))
        """
    )
    return CardstackIdentifier(res["version"], res["type"], res["uniqueId"])



import re

from Crypto.Hash import SHAKE128  # type: ignore


def hash_func(data: str):
    shake = SHAKE128.new()
    shake.update(str.encode(data))
    return shake.read(SHAKE_128_OUTPUT_BYTES).hex()


def to_kebab_case(camel_case: str):
    return re.sub(r"(?<!^)(?=[A-Z])", "-", camel_case).lower()


# 1 byte = 8 bits
# 1 byte = 2 hex digits
SHAKE_128_OUTPUT_BITS = 64
SHAKE_128_OUTPUT_BYTES = 8
HASH_LENGTH = int(SHAKE_128_OUTPUT_BITS / 4)

ASCII_START_ZERO = 48
ASCII_START_TEN = 65
ASCII_START_THIRTY_SIX = 97


class VersionOutOfRange(Exception):
    pass


def is_supported_version(char_code: int):
    return (
        (char_code >= ASCII_START_ZERO and char_code <= ASCII_START_ZERO + 9)
        or (char_code >= ASCII_START_TEN and char_code <= ASCII_START_TEN + 25)
        or (
            char_code >= ASCII_START_THIRTY_SIX
            and char_code <= ASCII_START_THIRTY_SIX + 25
        )
    )


def version_from_char(char: str) -> int:
    char_code = ord(char)
    if char_code >= ASCII_START_ZERO and char_code <= ASCII_START_ZERO + 9:
        return char_code - ASCII_START_ZERO
    elif char_code >= ASCII_START_TEN and char_code <= ASCII_START_TEN + 25:
        return char_code - ASCII_START_TEN + 10
    elif (
        char_code >= ASCII_START_THIRTY_SIX and char_code <= ASCII_START_THIRTY_SIX + 25
    ):
        return char_code - ASCII_START_THIRTY_SIX + 36
    else:
        raise VersionOutOfRange("Unsupported version character {v}")


def number_to_version_char(v: int) -> str:
    """inverse of version_from_char()"""
    if v >= 0 and v <= 9:
        return chr(v + ASCII_START_ZERO)
    elif v >= 10 and v <= 10 + 25:
        return chr(v + ASCII_START_TEN - 10)
    elif v >= 36 and v <= 36 + 25:
        return chr(v + ASCII_START_THIRTY_SIX - 36)
    else:
        raise VersionOutOfRange(f"version out of supported range: {v}")



#!/usr/bin/env python3

import re
from typing import Literal, TypedDict
from uuid import UUID

from did_resolver import DIDDocument, DIDResolutionResult, ParsedDID, Resolvable
from shortuuid import ShortUUID
from typing_extensions import NotRequired

from .utils import HASH_LENGTH, hash_func, number_to_version_char, to_kebab_case

# https://github.com/skorokithakis/shortuuid/issues/68
# shortuuid decode() sorts alphabets before using it for translation
# this is incompatible with shortuuid implementation in javascript
# we therefore overwrite the _alphabet attribute so the answers are the same in javascript
FLICKR_BASE58_ALPHABET = "123456789abcdefghijkmnopqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ"
su = ShortUUID(alphabet=FLICKR_BASE58_ALPHABET)
su._alphabet = list(FLICKR_BASE58_ALPHABET)

UUIDV5_NAMESPACE = (
    "1b671a64-40d5-491e-99b0-da01ff1f3341"  # this NAMESPACE is reserved for uuid5
)
CURRENT_VERSION = 1
CardstackIdentifierType = Literal[
    "PrepaidCardCustomization",
    "MerchantInfo",
    "SupplierInfo",
    "CardSpace",
    "RewardRule",
]
SHORT_TYPE = Literal["p", "m", "s", "c", "r"]


def did_type_to_short_type(cardstack_identifier: CardstackIdentifierType) -> SHORT_TYPE:
    if cardstack_identifier == "PrepaidCardCustomization":
        return "p"
    elif cardstack_identifier == "MerchantInfo":
        return "m"
    elif cardstack_identifier == "SupplierInfo":
        return "s"
    elif cardstack_identifier == "CardSpace":
        return "c"
    elif cardstack_identifier == "RewardRule":
        return "r"


def short_type_to_did_type(short_type: str) -> CardstackIdentifierType:
    """inverse of did_type_to_short_type()"""
    if short_type == "p":
        return "PrepaidCardCustomization"
    elif short_type == "m":
        return "MerchantInfo"
    elif short_type == "s":
        return "SupplierInfo"
    elif short_type == "c":
        return "CardSpace"
    elif short_type == "r":
        return "RewardRule"
    else:
        raise Exception(f'Invalid DID identifier: unknown type "{short_type}"')


class CardstackIdentifier:
    version: int
    type: CardstackIdentifierType
    uniqueId: str

    def __init__(
        self, version: int, type: CardstackIdentifierType, unique_id: str
    ) -> None:
        self.version = version
        self.type = type
        self.uniqueId = normalize_unique_id(unique_id)

    def to_did(self):
        version_string = number_to_version_char(self.version)
        result = f"{version_string}{did_type_to_short_type(self.type)}{self.uniqueId}"
        checksum = hash_func(result)
        return f"did:cardstack:{result}{checksum}"

    def __repr__(self):
        return f"""{self.__class__.__name__}(version: {self.version}, type: {self.type}, unique_id: {self.uniqueId})"""


def normalize_unique_id(candidate: str):
    if is_flickr_base_58(candidate):
        return candidate
    else:
        if is_valid_uuid(candidate):
            return su.encode(candidate)
        else:
            raise Exception(
                f'uniqueId must be a flickrBase58 or RFC4122 v4-compliant UUID. Was: "{candidate}"'
            )


def is_valid_uuid(uuid_to_test, version=4):
    try:
        UUID(uuid_to_test, version=version)
        return True
    except ValueError:
        return False


def is_flickr_base_58(candidate: str) -> bool:
    BASE_58_CHAR_LENGTH = 22
    return len(candidate) == BASE_58_CHAR_LENGTH and bool(
        re.match(f"^[{FLICKR_BASE58_ALPHABET}]+$", candidate)
    )


def s3_resolution_method(
    did: str, parsed_did: ParsedDID, resolver: Resolvable
) -> DIDResolutionResult:
    cardstack_identifier = parse_identifier(parsed_did["id"])
    path = to_kebab_case(cardstack_identifier.type)
    did_document: DIDDocument = {
        "@context": [
            "https://www.w3.org/ns/did/v1",
            "https://identity.foundation/EcdsaSecp256k1RecoverySignature2020/lds-ecdsa-secp256k1-recovery2020-0.0.jsonld",
        ],
        "id": did,
        "alsoKnownAs": [
            f"https://storage.cardstack.com/{path}/{cardstack_identifier.uniqueId}.json"
        ],
        "verificationMethod": [],
        "authentication": [],
        "assertionMethod": [],
    }
    return {
        "didResolutionMetadata": {"contentType": "application/did+ld+json"},
        "didDocument": did_document,
        "didDocumentMetadata": {},
    }


def parse_identifier(identifier: str) -> CardstackIdentifier:
    cutoff = len(identifier) - HASH_LENGTH
    data = identifier[:cutoff]
    checksum = identifier[cutoff : len(identifier)]
    if checksum != hash_func(data):
        raise Exception("Invalid DID identifier: checksum failed")
    version = data[0]
    type = short_type_to_did_type(data[1])
    unique_id = data[2:]
    return CardstackIdentifier(version, type, unique_id)


class EncodeOptions(TypedDict):
    type: CardstackIdentifierType
    version: NotRequired[int]
    uniqueId: NotRequired[str]


def encode_did(opts: EncodeOptions) -> str:
    version = opts.get("version", CURRENT_VERSION)
    unique_id = opts.get("uniqueId", su.uuid())
    return CardstackIdentifier(version, opts["type"], unique_id).to_did()


def get_resolver():
    return {"cardstack": s3_resolution_method}



import re
from datetime import timedelta
from typing import get_args

import pytest
from did_resolver.resolver import Resolver
from hypothesis import given, settings
from hypothesis import strategies as st
from python_did_resolver.main import (
    FLICKR_BASE58_ALPHABET,
    SHORT_TYPE,
    CardstackIdentifierType,
    encode_did,
    hash_func,
    is_flickr_base_58,
    is_valid_uuid,
    parse_identifier,
    s3_resolution_method,
    su,
)
from python_did_resolver.utils import VersionOutOfRange, to_kebab_case

from .call_js_resolver import js_encode_did, js_parse_identifier, js_resolve_did

encode_options = st.fixed_dictionaries(
    {
        "type": st.sampled_from(
            [
                "PrepaidCardCustomization",
                "MerchantInfo",
                "SupplierInfo",
                "CardSpace",
                "RewardRule",
            ]
        ),
    },
    optional={
        "version": st.integers(min_value=1, max_value=61),
        "uniqueId": st.uuids().map(lambda uuid: su.encode(uuid)),
    },
)

encode_options_bad_unique_id = st.fixed_dictionaries(
    {
        "type": st.sampled_from(
            [
                "PrepaidCardCustomization",
                "MerchantInfo",
                "SupplierInfo",
                "CardSpace",
                "RewardRule",
            ]
        ),
        "uniqueId": st.sampled_from(["foo"]),
    },
    optional={"version": st.integers(min_value=1, max_value=61)},
)
encode_options_bad_version = st.fixed_dictionaries(
    {
        "type": st.sampled_from(
            [
                "PrepaidCardCustomization",
                "MerchantInfo",
                "SupplierInfo",
                "CardSpace",
                "RewardRule",
            ]
        ),
        "version": st.one_of(st.integers(max_value=-1), st.integers(min_value=62)),
    },
    optional={
        "uniqueId": st.uuids().map(lambda uuid: su.encode(uuid)),
    },
)


@given(st.uuids())
def test_shortuuid(o):
    """
    This test ensures that the shortuuid class is patched correctly
    """
    assert su._alphabet == list(FLICKR_BASE58_ALPHABET)
    o == su.decode(su.encode(o))


@given(encode_options)
@settings(deadline=timedelta(seconds=60))
def test_encode(o):
    did = encode_did(o)
    assert isinstance(did, str)
    assert bool(re.match("^did:cardstack:", did))
    identifier = did.split(":")[2]
    parsed = parse_identifier(identifier)
    assert isinstance(parsed.uniqueId, str)
    assert is_valid_uuid(str(su.decode(parsed.uniqueId)), version=4) or is_valid_uuid(
        str(su.decode(parsed.uniqueId)), version=5
    )
    assert is_flickr_base_58(parsed.uniqueId)
    assert parsed.type in get_args(CardstackIdentifierType)  # runtime check


@given(encode_options)
@settings(deadline=timedelta(seconds=60))
def test_encode_js(o):
    did = js_encode_did(o)
    assert isinstance(did, str)
    assert bool(re.match("^did:cardstack:", did))
    identifier = did.split(":")[2]
    parsed = js_parse_identifier(identifier)
    assert isinstance(parsed.uniqueId, str)
    assert is_valid_uuid(str(su.decode(parsed.uniqueId)), version=4) or is_valid_uuid(
        str(su.decode(parsed.uniqueId)), version=5
    )
    assert is_flickr_base_58(parsed.uniqueId)
    assert parsed.type in get_args(CardstackIdentifierType)  # runtime check


@given(encode_options_bad_unique_id)
def test_encode_bad_unique_id(o):
    with pytest.raises(
        Exception,
        match=r"uniqueId must be a flickrBase58 or RFC4122 v4-compliant UUID. Was: .*",
    ):
        encode_did(o)


@given(encode_options_bad_version)
@settings(deadline=timedelta(seconds=60))
def test_encode_bad_version(o):
    with pytest.raises(VersionOutOfRange):
        encode_did(o)


@given(encode_options)
@settings(deadline=timedelta(seconds=60))
def test_compat_encode(o):
    did = encode_did(o)
    did_js = js_encode_did(o)
    did == did_js


@given(encode_options)
def test_resolve(o):
    did = encode_did(o)
    path = to_kebab_case(o["type"])
    registry = {"cardstack": s3_resolution_method}
    r = Resolver(registry)
    resolved_doc = r.resolve(did)
    if "alsoKnownAs" in resolved_doc["didDocument"].keys() and "unique_id" in o.keys():
        uid = o["unique_id"]
        assert (
            resolved_doc["didDocument"]["alsoKnownAs"][0]
            == f"https://storage.cardstack.com/{path}/{uid}.json"
        )


@given(encode_options)
def test_resolve_invalid_checksum(o):
    did = encode_did(o) + "a"
    registry = {"cardstack": s3_resolution_method}
    r = Resolver(registry)
    with pytest.raises(
        Exception,
        match=r"Invalid DID identifier: checksum failed",
    ):
        r.resolve(did)


SHORT_TYPES_DELIMITED = "".join([str(ele) + "|" for ele in get_args(SHORT_TYPE)])
REGEX_NON_EXISTENT_SHORT_CODE = rf"^(?!(?:{SHORT_TYPES_DELIMITED})$)[a-z]\Z"


@given(st.from_regex(REGEX_NON_EXISTENT_SHORT_CODE))
def test_resolve_invalid_short_type(c):
    cardstack_identifier = "1" + c + su.random(length=22)
    did = f"did:cardstack:{cardstack_identifier}{hash_func(cardstack_identifier)}"
    registry = {"cardstack": s3_resolution_method}
    r = Resolver(registry)
    with pytest.raises(
        Exception,
        match=r"Invalid DID identifier: unknown type .*",
    ):
        r.resolve(did)


@given(encode_options)
@settings(deadline=timedelta(seconds=60))
def test_resolve_js(o):
    did = js_encode_did(o)
    path = to_kebab_case(o["type"])
    resolved_doc = js_resolve_did(did)
    if "alsoKnownAs" in resolved_doc["didDocument"].keys() and "unique_id" in o.keys():
        uid = o["unique_id"]
        assert (
            resolved_doc["didDocument"]["alsoKnownAs"][0]
            == f"https://storage.cardstack.com/{path}/{uid}.json"
        )


@given(encode_options)
@settings(deadline=timedelta(seconds=60))
def test_compat_resolve(o):
    did = encode_did(o)
    registry = {"cardstack": s3_resolution_method}
    r = Resolver(registry)
    r.resolve(did) == js_resolve_did(did)
